@@ -19,14 +19,12 @@
 
 #include "config.h"
 
-/* Allow usage of tracker_namespace_manager_get_default() */
-#define TRACKER_VERSION_MIN_REQUIRED TRACKER_VERSION_3_2
-
 #include <glib.h>
 #include <json-glib/json-glib.h>
 
 #include <string.h>
 
+#include <tracker-deserializer-resource.h>
 #include <tracker-uri.h>
 #include <tracker-resource.h>
 #include <tracker-ontologies.h>
@@ -35,7 +33,7 @@
 #include "tracker-utils.h"
 
 /* For prefixed names parsing */
-#include "libtracker-data/tracker-sparql-grammar.h"
+#include <libtracker-sparql/core/tracker-sparql-grammar.h>
 
 #include <tracker-private.h>
 
@@ -81,7 +79,6 @@ enum {
 };
 
 static void dispose      (GObject *object);
-static void constructed  (GObject *object);
 static void finalize     (GObject *object);
 static void get_property (GObject    *object,
                           guint       param_id,
@@ -92,6 +89,47 @@ static void set_property (GObject      *object,
                           const GValue *value,
                           GParamSpec   *pspec);
 
+static char *
+escape_iri (const gchar *str)
+{
+	GString *iri;
+
+	/* Escapes IRI references according to IRIREF in SPARQL grammar definition,
+	 * further validation on IRI validity may happen deeper down.
+	 */
+
+	if (!str)
+		return NULL;
+
+	/* Fast path, check whether there's no characters to escape */
+	if (!strpbrk (str,
+	              "<>\"{}|^`"
+	              "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10"
+	              "\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x20")) {
+		return g_strdup (str);
+	}
+
+	iri = g_string_new (NULL);
+
+	while (*str != '\0') {
+		gunichar unichar;
+
+		unichar = g_utf8_get_char (str);
+		str = g_utf8_next_char (str);
+
+		if (unichar <= 0x20 ||
+		    unichar == '<' || unichar == '>' ||
+		    unichar == '"' || unichar == '{' ||
+		    unichar == '}' || unichar == '|' ||
+		    unichar == '^' || unichar == '`' ||
+		    unichar == '\\')
+			g_string_append_printf (iri, "%%%X", unichar);
+		else
+			g_string_append_unichar (iri, unichar);
+	}
+
+	return g_string_free (iri, FALSE);
+}
 
 static void
 tracker_resource_class_init (TrackerResourceClass *klass)
@@ -99,7 +137,6 @@ tracker_resource_class_init (TrackerResourceClass *klass)
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
 	object_class->dispose      = dispose;
-	object_class->constructed  = constructed;
 	object_class->finalize     = finalize;
 	object_class->get_property = get_property;
 	object_class->set_property = set_property;
@@ -160,20 +197,6 @@ dispose (GObject *object)
 	G_OBJECT_CLASS (tracker_resource_parent_class)->dispose (object);
 }
 
-static void
-constructed (GObject *object)
-{
-	TrackerResourcePrivate *priv;
-
-	priv = GET_PRIVATE (TRACKER_RESOURCE (object));
-
-	if (! priv->identifier) {
-		priv->identifier = generate_blank_node_identifier ();
-	}
-
-	G_OBJECT_CLASS (tracker_resource_parent_class)->constructed (object);
-}
-
 static void
 finalize (GObject *object)
 {
@@ -181,9 +204,7 @@ finalize (GObject *object)
 
 	priv = GET_PRIVATE (TRACKER_RESOURCE (object));
 
-	if (priv->identifier) {
-		g_free (priv->identifier);
-	}
+	g_clear_pointer (&priv->identifier, g_free);
 
 	(G_OBJECT_CLASS (tracker_resource_parent_class)->finalize)(object);
 }
@@ -194,13 +215,9 @@ get_property (GObject    *object,
               GValue     *value,
               GParamSpec *pspec)
 {
-	TrackerResourcePrivate *priv;
-
-	priv = GET_PRIVATE (TRACKER_RESOURCE (object));
-
 	switch (param_id) {
 	case PROP_IDENTIFIER:
-		g_value_set_string (value, priv->identifier);
+		g_value_set_string (value, tracker_resource_get_identifier (TRACKER_RESOURCE (object)));
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
@@ -326,6 +343,13 @@ validate_pointer (const void *pointer,
 	return TRUE;
 }
 
+static void
+value_set_uri (GValue      *value,
+               const gchar *uri)
+{
+	g_value_take_string (value, escape_iri (uri));
+}
+
 #define SET_PROPERTY_FOR_GTYPE(name, ctype, gtype, set_function, validate_function) \
 	void name (TrackerResource *self,                                           \
 	           const char *property_uri,                                        \
@@ -442,7 +466,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_string, const char *, G_TYPE_STRING
  * produces similar RDF to tracker_resource_set_relation(), although
  * it requires that the URI is previously known.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer)
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_uri, const char *, TRACKER_TYPE_URI, value_set_uri, validate_pointer)
 
 /**
  * tracker_resource_set_datetime:
@@ -656,7 +680,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_string, const char *, G_TYPE_STRING
  * produces similar RDF to tracker_resource_add_relation(), although
  * it requires that the URI is previously known.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer)
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_uri, const char *, TRACKER_TYPE_URI, value_set_uri, validate_pointer)
 
 /**
  * tracker_resource_add_datetime:
@@ -854,6 +878,9 @@ tracker_resource_get_identifier (TrackerResource *self)
 
 	priv = GET_PRIVATE (self);
 
+	if (!priv->identifier)
+		priv->identifier = generate_blank_node_identifier ();
+
 	return priv->identifier;
 }
 
@@ -880,19 +907,8 @@ tracker_resource_set_identifier (TrackerResource *self,
 
 	priv = GET_PRIVATE (self);
 
-	g_free (priv->identifier);
-
-	if (identifier == NULL) {
-		/* We take NULL to mean "this is a blank node", and generate a
-		 * unique blank node identifier right away. This is easier than
-		 * leaving it NULL and trying to generate a blank node ID at
-		 * serialization time, and it means that the serialization
-		 * output is stable when called multiple times.
-		 */
-		priv->identifier = generate_blank_node_identifier ();
-	} else {
-		priv->identifier = g_strdup (identifier);
-	}
+	g_clear_pointer (&priv->identifier, g_free);
+	priv->identifier = escape_iri (identifier);
 }
 
 /**
@@ -910,14 +926,10 @@ gint
 tracker_resource_identifier_compare_func (TrackerResource *resource,
                                           const char      *identifier)
 {
-	TrackerResourcePrivate *priv;
-
 	g_return_val_if_fail (TRACKER_IS_RESOURCE (resource), 0);
 	g_return_val_if_fail (identifier != NULL, 0);
 
-	priv = GET_PRIVATE (resource);
-
-	return strcmp (priv->identifier, identifier);
+	return strcmp (tracker_resource_get_identifier (resource), identifier);
 }
 
 /**
@@ -937,15 +949,11 @@ gint
 tracker_resource_compare (TrackerResource *a,
                           TrackerResource *b)
 {
-	TrackerResourcePrivate *a_priv, *b_priv;
-
 	g_return_val_if_fail (TRACKER_IS_RESOURCE (a), 0);
 	g_return_val_if_fail (TRACKER_IS_RESOURCE (b), 0);
 
-	a_priv = GET_PRIVATE (a);
-	b_priv = GET_PRIVATE (b);
-
-	return strcmp (a_priv->identifier, b_priv->identifier);
+	return strcmp (tracker_resource_get_identifier (a),
+		       tracker_resource_get_identifier (b));
 }
 
 /**
@@ -1027,14 +1035,6 @@ maybe_intern_prefix_of_compact_uri (TrackerNamespaceManager *all_namespaces,
 }
 
 
-typedef struct {
-	TrackerNamespaceManager *all_namespaces, *our_namespaces;
-	GString *string;
-	GList *done_list;
-} GenerateTurtleData;
-
-static void generate_turtle (TrackerResource *resource, GenerateTurtleData *data);
-
 static gboolean
 is_blank_node (const char *uri_or_curie_or_blank)
 {
@@ -1065,50 +1065,6 @@ is_builtin_class (const gchar             *uri_or_curie,
 	return has_prefix;
 }
 
-static void
-generate_nested_turtle_resource (TrackerResource    *resource,
-                                 GenerateTurtleData *data)
-{
-	/* We don't need to produce turtle for builtin classes */
-	if (is_builtin_class (tracker_resource_get_identifier (resource),
-	                      data->all_namespaces))
-		return;
-
-	if (g_list_find (data->done_list, resource) == NULL) {
-		data->done_list = g_list_prepend (data->done_list, resource);
-		generate_turtle (resource, data);
-		g_string_append (data->string, "\n");
-	}
-}
-
-static void
-generate_turtle_resources_foreach (gpointer key,
-                                   gpointer value_ptr,
-                                   gpointer user_data)
-{
-	const GValue *value = value_ptr;
-	GenerateTurtleData *data = user_data;
-	TrackerResource *resource;
-	guint i;
-
-	if (G_VALUE_HOLDS (value, TRACKER_TYPE_RESOURCE)) {
-		resource = g_value_get_object (value);
-		generate_nested_turtle_resource (resource, data);
-	} else if (G_VALUE_HOLDS (value, G_TYPE_PTR_ARRAY)) {
-		GPtrArray *array = g_value_get_boxed (value);
-		const GValue *array_value;
-
-		for (i = 0; i < array->len; i++) {
-			array_value = g_ptr_array_index (array, i);
-
-			if (G_VALUE_HOLDS (array_value, TRACKER_TYPE_RESOURCE)) {
-				resource = g_value_get_object (array_value);
-				generate_nested_turtle_resource (resource, data);
-			}
-		}
-	}
-}
-
 static void
 generate_turtle_uri_value (const char              *uri_or_curie_or_blank,
                            GString                 *string,
@@ -1232,39 +1188,6 @@ generate_turtle_property (const char              *property,
 	}
 }
 
-static void
-generate_turtle (TrackerResource    *resource,
-                 GenerateTurtleData *data)
-{
-	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
-	GHashTableIter iter;
-	const char *property;
-	const GValue *value;
-
-	/* First we recurse to any relations that aren't already in the done list */
-	g_hash_table_foreach (priv->properties, generate_turtle_resources_foreach, data);
-
-	generate_turtle_uri_value (tracker_resource_get_identifier(resource),
-	        data->string, data->all_namespaces, data->our_namespaces);
-	g_string_append (data->string, " ");
-
-	g_hash_table_iter_init (&iter, priv->properties);
-	if (g_hash_table_iter_next (&iter, (gpointer *)&property, (gpointer *)&value)) {
-		while (TRUE) {
-			generate_turtle_property (property, value, data->string, data->all_namespaces, data->our_namespaces);
-
-			maybe_intern_prefix_of_compact_uri (data->all_namespaces, data->our_namespaces, property);
-
-			if (g_hash_table_iter_next (&iter, (gpointer *)&property, (gpointer *)&value)) {
-				g_string_append (data->string, " ;\n  ");
-			} else {
-				g_string_append (data->string, " .\n");
-				break;
-			}
-		}
-	}
-}
-
 /**
  * tracker_resource_print_turtle:
  * @self: a #TrackerResource
@@ -1281,46 +1204,22 @@ generate_turtle (TrackerResource    *resource,
  * from the connection that is the intended recipient of this data.
  *
  * Returns: a newly-allocated string
+ *
+ * Deprecated: 3.4: Use tracker_resource_print_rdf() instead.
  */
 char *
 tracker_resource_print_turtle (TrackerResource         *self,
                                TrackerNamespaceManager *namespaces)
 {
-	TrackerResourcePrivate *priv;
-	GenerateTurtleData context;
-	char *prefixes;
-
 	g_return_val_if_fail (TRACKER_IS_RESOURCE (self), "");
 
-	priv = GET_PRIVATE (self);
-
 	if (namespaces == NULL) {
+		G_GNUC_BEGIN_IGNORE_DEPRECATIONS
 		namespaces = tracker_namespace_manager_get_default ();
+		G_GNUC_END_IGNORE_DEPRECATIONS
 	}
 
-	if (g_hash_table_size (priv->properties) == 0) {
-		return g_strdup("");
-	}
-
-	context.all_namespaces = namespaces;
-	context.our_namespaces = tracker_namespace_manager_new ();
-	context.string = g_string_new ("");
-	context.done_list = g_list_prepend (NULL, self);
-
-	maybe_intern_prefix_of_compact_uri (context.all_namespaces, context.our_namespaces, tracker_resource_get_identifier(self));
-
-	generate_turtle (self, &context);
-
-	prefixes = tracker_namespace_manager_print_turtle (context.our_namespaces);
-	g_string_prepend (context.string, "\n");
-	g_string_prepend (context.string, prefixes);
-
-	g_object_unref (context.our_namespaces);
-	g_free (prefixes);
-
-	g_list_free (context.done_list);
-
-	return g_string_free (context.string, FALSE);
+	return tracker_resource_print_rdf (self, namespaces, TRACKER_RDF_FORMAT_TURTLE, NULL);
 }
 
 typedef struct {
@@ -1434,7 +1333,8 @@ generate_sparql_delete_queries (TrackerResource     *resource,
 			}
 
 			g_string_append (data->string, "  ");
-			generate_turtle_uri_value (priv->identifier, data->string, data->namespaces, NULL);
+			generate_turtle_uri_value (tracker_resource_get_identifier (resource),
+						   data->string, data->namespaces, NULL);
 			g_string_append_printf (data->string, " %s ?%s }", property, variable_name);
 			g_free (variable_name);
 
@@ -1459,7 +1359,7 @@ generate_sparql_deletes (TrackerResource    *resource,
 
 	data->done_list = g_list_prepend (data->done_list, resource);
 
-	if (! is_blank_node (priv->identifier) && g_hash_table_size (priv->overwrite) > 0) {
+	if (!tracker_resource_is_blank_node (resource) && g_hash_table_size (priv->overwrite) > 0) {
 		generate_sparql_delete_queries (resource, priv->overwrite, data);
 	}
 
@@ -1487,7 +1387,8 @@ generate_sparql_insert_pattern (TrackerResource    *resource,
 	/* First, emit any sub-resources. */
 	g_hash_table_foreach (priv->properties, generate_sparql_relation_inserts_foreach, data);
 
-	generate_turtle_uri_value (priv->identifier, data->string, data->namespaces, NULL);
+	generate_turtle_uri_value (tracker_resource_get_identifier (resource),
+				   data->string, data->namespaces, NULL);
 	g_string_append_printf (data->string, " ");
 
 	/* rdf:type needs to be first, otherwise you'll see 'subject x is not in domain y'
@@ -1549,7 +1450,9 @@ tracker_resource_print_sparql_update (TrackerResource         *resource,
 	priv = GET_PRIVATE(resource);
 
 	if (namespaces == NULL) {
+		G_GNUC_BEGIN_IGNORE_DEPRECATIONS
 		namespaces = tracker_namespace_manager_get_default ();
+		G_GNUC_END_IGNORE_DEPRECATIONS
 	}
 
 	if (g_hash_table_size (priv->properties) == 0) {
@@ -1616,7 +1519,7 @@ tracker_resource_generate_jsonld (TrackerResource    *self,
 	 * (where the caller passed NULL as an identifier) than to emit something
 	 * SPARQL-specific like '_:123'.
 	 */
-	if (strncmp (priv->identifier, "_:", 2) != 0) {
+	if (!tracker_resource_is_blank_node (self)) {
 		json_builder_set_member_name (builder, "@id");
 		json_builder_add_string_value (builder, priv->identifier);
 	}
@@ -1733,7 +1636,9 @@ tracker_resource_print_jsonld (TrackerResource         *self,
 	char *result;
 
 	if (namespaces == NULL) {
+		G_GNUC_BEGIN_IGNORE_DEPRECATIONS
 		namespaces = tracker_namespace_manager_get_default ();
+		G_GNUC_END_IGNORE_DEPRECATIONS
 	}
 
 	context.all_namespaces = namespaces;
@@ -1770,6 +1675,87 @@ tracker_resource_print_jsonld (TrackerResource         *self,
 	return result;
 }
 
+static TrackerSerializerFormat
+convert_format (TrackerRdfFormat format)
+{
+	switch (format) {
+	case TRACKER_RDF_FORMAT_TURTLE:
+		return TRACKER_SERIALIZER_FORMAT_TTL;
+	case TRACKER_RDF_FORMAT_TRIG:
+		return TRACKER_SERIALIZER_FORMAT_TRIG;
+	case TRACKER_N_RDF_FORMATS:
+		g_assert_not_reached ();
+	}
+
+	return -1;
+}
+
+/**
+ * tracker_resource_print_rdf:
+ * @self: a #TrackerResource
+ * @namespaces: a set of prefixed URLs
+ * @format: RDF format of the printed string
+ * @graph: target graph of the resource RDF, or %NULL for the default graph
+ *
+ * Serialize all the information in @resource into the selected RDF format.
+ *
+ * The @namespaces object is used to expand any compact URI values. In most
+ * cases you should pass the one returned by tracker_sparql_connection_get_namespace_manager()
+ * from the connection that is the intended recipient of this data.
+ *
+ * Returns: a newly-allocated string containing RDF data in the requested format.
+ *
+ * Since: 3.4
+ **/
+char *
+tracker_resource_print_rdf (TrackerResource         *self,
+                            TrackerNamespaceManager *namespaces,
+                            TrackerRdfFormat         format,
+                            const gchar             *graph)
+{
+	TrackerSparqlCursor *deserializer;
+	GInputStream *serializer;
+	GString *str;
+
+	g_return_val_if_fail (TRACKER_IS_RESOURCE (self), NULL);
+	g_return_val_if_fail (TRACKER_IS_NAMESPACE_MANAGER (namespaces), NULL);
+	g_return_val_if_fail (format < TRACKER_N_RDF_FORMATS, NULL);
+
+#define BUF_SIZE 4096
+	deserializer = tracker_deserializer_resource_new (self, namespaces, graph);
+	serializer = tracker_serializer_new (TRACKER_SPARQL_CURSOR (deserializer),
+	                                     namespaces,
+	                                     convert_format (format));
+	g_object_unref (deserializer);
+
+	str = g_string_new (NULL);
+
+	while (TRUE) {
+		GBytes *bytes;
+
+		bytes = g_input_stream_read_bytes (serializer, BUF_SIZE, NULL, NULL);
+		if (!bytes) {
+			g_string_free (str, TRUE);
+			return NULL;
+		}
+
+		if (g_bytes_get_size (bytes) == 0) {
+			g_bytes_unref (bytes);
+			break;
+		}
+
+		g_string_append_len (str,
+		                     g_bytes_get_data (bytes, NULL),
+		                     g_bytes_get_size (bytes));
+		g_bytes_unref (bytes);
+	}
+#undef BUF_SIZE
+
+	g_object_unref (serializer);
+
+	return g_string_free (str, FALSE);
+}
+
 static GVariant *
 tracker_serialize_single_value (TrackerResource         *resource,
                                 const GValue            *value)
@@ -1823,8 +1809,7 @@ tracker_resource_serialize (TrackerResource *resource)
 
 	g_variant_builder_init (&builder, G_VARIANT_TYPE_ARRAY);
 
-	if (priv->identifier &&
-	    strncmp (priv->identifier, "_:", 2) != 0) {
+	if (!tracker_resource_is_blank_node (resource)) {
 		g_variant_builder_add (&builder, "{sv}", "@id",
 		                       g_variant_new_string (priv->identifier));
 	}
@@ -2016,3 +2001,69 @@ tracker_resource_get_property_overwrite (TrackerResource *resource,
 
 	return g_hash_table_contains (priv->overwrite, property_uri);
 }
+
+void
+tracker_resource_iterator_init (TrackerResourceIterator *iter,
+                                TrackerResource         *resource)
+{
+	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
+
+	bzero (iter, sizeof (TrackerResourceIterator));
+	g_hash_table_iter_init (&iter->prop_iter, priv->properties);
+}
+
+gboolean
+tracker_resource_iterator_next (TrackerResourceIterator  *iter,
+                                const gchar             **property,
+                                const GValue            **value)
+{
+	gpointer key, val;
+
+	if (iter->cur_values && iter->cur_prop) {
+		iter->idx++;
+
+		if (iter->idx < iter->cur_values->len) {
+			*property = iter->cur_prop;
+			*value = g_ptr_array_index (iter->cur_values, iter->idx);
+			return TRUE;
+		} else {
+			iter->cur_values = NULL;
+			iter->cur_prop = NULL;
+		}
+	}
+
+	if (!g_hash_table_iter_next (&iter->prop_iter, &key, &val))
+		return FALSE;
+
+	if (G_VALUE_HOLDS (val, G_TYPE_PTR_ARRAY)) {
+		iter->cur_prop = key;
+		iter->cur_values = g_value_get_boxed (val);
+		iter->idx = 0;
+		*property = iter->cur_prop;
+		*value = g_ptr_array_index (iter->cur_values, iter->idx);
+		return TRUE;
+	}
+
+	*property = key;
+	*value = val;
+	return TRUE;
+}
+
+const gchar *
+tracker_resource_get_identifier_internal (TrackerResource *resource)
+{
+	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
+
+	return priv->identifier;
+}
+
+gboolean
+tracker_resource_is_blank_node (TrackerResource *resource)
+{
+	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
+
+	if (!priv->identifier)
+		return TRUE;
+
+	return strncmp (priv->identifier, "_:", 2) == 0;
+}
